# This is a port of the ruby zabbix api found here:
# http://trac.red-tux.net/browser/ruby/api/zbx_api.rb
#
#LGPL 2.1   http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html
#Zabbix API Python Library.
#Original Ruby Library is Copyright (C) 2009 Andrew Nelson nelsonab(at)red-tux(dot)net
#Python Library is Copyright (C) 2009 Brett Lentz brett.lentz(at)gmail(dot)com
#
#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU Lesser General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.
#
#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#Lesser General Public License for more details.
#
#You should have received a copy of the GNU Lesser General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA


# NOTES:
# The API requires zabbix 1.8 or later.
# Currently, not all of the API is implemented, and some functionality is
# broken. This is a work in progress.

import base64
import hashlib
import logging
import string
import sys
import re
import datetime
try:
    import urllib2
except ImportError:
    import urllib.request as urllib2  # python3
from collections import deque

__logger = logging.getLogger(__name__)

try:
    # Separate module or Python <2.6
    import simplejson as json
    __logger.log(15, "Using simplejson library")
except ImportError:
    # Python >=2.6
    import json
    __logger.log(15, "Using native json library")


def checkauth(fn):
    """ Decorator to check authentication of the decorated method """
    def ret(self, *args):
        self.__checkauth__()
        return fn(self, args)
    return ret


def dojson(name):
    def decorator(fn):
        def wrapper(self, opts):
            self.logger.log(logging.DEBUG, \
                    "Going to do_request for %s with opts %s" \
                    % (repr(fn), repr(opts)))
            return self.do_request(self.json_obj(name, opts))['result']
        return wrapper
    return decorator


def dojson2(fn):
    def wrapper(self, method, opts):
        self.logger.log(logging.DEBUG, \
                "Going to do_request for %s with opts %s" \
                % (repr(fn), repr(opts)))
        return self.do_request(self.json_obj(method, opts))['result']
    return wrapper


class ZabbixAPIException(Exception):
    """ generic zabbix api exception
    code list:
         -32602 - Invalid params (eg already exists)
         -32500 - no permissions
    """
    pass


class Already_Exists(ZabbixAPIException):
    pass


class InvalidProtoError(ZabbixAPIException):
    """ Recived an invalid proto """
    pass


class ZabbixAPI(object):
    __username__ = ''
    __password__ = ''

    auth = ''
    url = '/api_jsonrpc.php'
    params = None
    method = None
    # HTTP or HTTPS
    proto = 'http'
    # HTTP authentication
    httpuser = None
    httppasswd = None
    timeout = 10
    # sub-class instances.
    user = None
    usergroup = None
    host = None
    item = None
    hostgroup = None
    hostinterface = None
    application = None
    trigger = None
    sysmap = None
    template = None
    drule = None
    # Constructor Params:
    # server: Server to connect to
    # path: Path leading to the zabbix install
    # proto: Protocol to use. http or https
    # We're going to use proto://server/path to find the JSON-RPC api.
    #
    # user: HTTP auth username
    # passwd: HTTP auth password
    # log_level: logging level
    # r_query_len: max len query history
    # **kwargs: Data to pass to each api module

    TRIGGER_SEVERITY = {
            0: 'not_classified',
            1: 'information',
            2: 'warning',
            3: 'average',
            4: 'high',
            5: 'disaster',
    }

    def __init__(self, server='http://localhost/zabbix', user=httpuser, passwd=httppasswd,
                 log_level=logging.WARNING, timeout=10, r_query_len=10, **kwargs):
        """ Create an API object.  """
        self._setuplogging()
        self.set_log_level(log_level)
        self.server = server
        self.url = server + '/api_jsonrpc.php'
        self.proto = self.server.split("://")[0]
        #self.proto=proto
        self.httpuser = user
        self.httppasswd = passwd
        self.timeout = timeout
        self.usergroup = ZabbixAPISubClass(self, dict({"prefix": "usergroup"}, **kwargs))
        self.user = ZabbixAPISubClass(self, dict({"prefix": "user"}, **kwargs))
        self.host = ZabbixAPISubClass(self, dict({"prefix": "host"}, **kwargs))
        self.item = ZabbixAPISubClass(self, dict({"prefix": "item"}, **kwargs))
        self.hostgroup = ZabbixAPISubClass(self, dict({"prefix": "hostgroup"}, **kwargs))
        self.hostinterface = ZabbixAPISubClass(self, dict({"prefix": "hostinterface"}, **kwargs))
        self.application = ZabbixAPISubClass(self, dict({"prefix": "application"}, **kwargs))
        self.trigger = ZabbixAPISubClass(self, dict({"prefix": "trigger"}, **kwargs))
        self.template = ZabbixAPISubClass(self, dict({"prefix": "template"}, **kwargs))
        self.action = ZabbixAPISubClass(self, dict({"prefix": "action"}, **kwargs))
        self.alert = ZabbixAPISubClass(self, dict({"prefix": "alert"}, **kwargs))
        self.info = ZabbixAPISubClass(self, dict({"prefix": "info"}, **kwargs))
        self.event = ZabbixAPISubClass(self, dict({"prefix": "event"}, **kwargs))
        self.graph = ZabbixAPISubClass(self, dict({"prefix": "graph"}, **kwargs))
        self.graphitem = ZabbixAPISubClass(self, dict({"prefix": "graphitem"}, **kwargs))
        self.map = ZabbixAPISubClass(self, dict({"prefix": "map"}, **kwargs))
        self.screen = ZabbixAPISubClass(self, dict({"prefix": "screen"}, **kwargs))
        self.script = ZabbixAPISubClass(self, dict({"prefix": "script"}, **kwargs))
        self.usermacro = ZabbixAPISubClass(self, dict({"prefix": "usermacro"}, **kwargs))
        self.drule = ZabbixAPISubClass(self, dict({"prefix": "drule"}, **kwargs))
        self.history = ZabbixAPISubClass(self, dict({"prefix": "history"}, **kwargs))
        self.maintenance = ZabbixAPISubClass(self, dict({"prefix": "maintenance"}, **kwargs))
        self.proxy = ZabbixAPISubClass(self, dict({"prefix": "proxy"}, **kwargs))
        self.apiinfo = ZabbixAPISubClass(self, dict({"prefix": "apiinfo"}, **kwargs))
        self.configuration = ZabbixAPISubClass(self, dict({"prefix": "configuration"}, **kwargs))
        self.dcheck = ZabbixAPISubClass(self, dict({"prefix": "dcheck"}, **kwargs))
        self.dhost = ZabbixAPISubClass(self, dict({"prefix": "dhost"}, **kwargs))
        self.discoveryrule = ZabbixAPISubClass(self, dict({"prefix": "discoveryrule"}, **kwargs))
        self.dservice = ZabbixAPISubClass(self, dict({"prefix": "dservice"}, **kwargs))
        self.iconmap = ZabbixAPISubClass(self, dict({"prefix": "iconmap"}, **kwargs))
        self.image = ZabbixAPISubClass(self, dict({"prefix": "image"}, **kwargs))
        self.mediatype = ZabbixAPISubClass(self, dict({"prefix": "mediatype"}, **kwargs))
        self.service = ZabbixAPISubClass(self, dict({"prefix": "service"}, **kwargs))
        self.templatescreen = ZabbixAPISubClass(self, dict({"prefix": "templatescreen"}, **kwargs))
        self.usermedia = ZabbixAPISubClass(self, dict({"prefix": "usermedia"}, **kwargs))
        self.hostinterface = ZabbixAPISubClass(self, dict({"prefix": "hostinterface"}, **kwargs))
        self.triggerprototype = ZabbixAPISubClass(self, dict({"prefix": "triggerprototype"}, **kwargs))
        self.graphprototype = ZabbixAPISubClass(self, dict({"prefix": "graphprototype"}, **kwargs))
        self.itemprototype = ZabbixAPISubClass(self, dict({"prefix": "itemprototype"}, **kwargs))
        self.webcheck = ZabbixAPISubClass(self, dict({"prefix": "webcheck"}, **kwargs))
        self.id = 0
        self.r_query = deque([], maxlen=r_query_len)
        self.debug(logging.INFO, "url: " + self.url)

    def _setuplogging(self):
        self.logger = logging.getLogger("zabbix_api.%s" % self.__class__.__name__)

    def set_log_level(self, level):
        self.debug(logging.INFO, "Set logging level to %d" % level)
        self.logger.setLevel(level)

    def get_severity(self, prio):
        """
        Return severity string from severity id.
        """
        try:
            return self.TRIGGER_SEVERITY[int(prio)]
        except KeyError:
            return 'unknown'

    def get_datetime(self, timestamp):
        """
        Return python datetime object from unix timestamp.
        """
        return datetime.datetime.fromtimestamp(int(timestamp))

    def convert_datetime(self, dt, format='%Y-%m-%d %H:%M:%S'):
        """
        Convert python datetime to human readable date and time string.
        """
        return dt.strftime(format)

    def get_age(self, dt):
        """
        Calculate delta between current time and datetime and return a
        human readable form of the delta object.
        """
        delta = datetime.datetime.now() - dt
        days = delta.days
        hours, rem = divmod(delta.seconds, 3600)
        minutes, seconds = divmod(rem, 60)
        if days:
            return '%dd %dh %dm' % (days, hours, minutes)
        else:
            return '%dh %dm %ds' % (hours, minutes, seconds)

    def recent_query(self):
        """
        return recent query
        """
        return list(self.r_query)

    def debug(self, level, var="", msg=None, safe=False):
        if msg:
            strval = str(msg)
        if var != "":
            strval = str(var)

        if safe:
            # Do not log passwords and tokens
            strval = re.sub(r'("password": )".*?"', r'\1"***"', strval)
            strval = re.sub(r'("auth": )".*?"', r'\1"***"', strval)

        self.logger.log(level, strval)

    def json_obj(self, method, params={}):
        obj = {'jsonrpc': '2.0',
               'method': method,
               'params': params,
               'auth': self.auth,
               'id': self.id
              }

        self.debug(logging.DEBUG, "json_obj: " + str(obj))

        return json.dumps(obj)

    def login(self, user='', password='', save=True):
        if user != '':
            l_user = user
            l_password = password

            if save:
                self.__username__ = user
                self.__password__ = password
        elif self.__username__ != '':
            l_user = self.__username__
            l_password = self.__password__
        else:
            raise ZabbixAPIException("No authentication information available.")

        # don't print the raw password.
        hashed_pw_string = "md5(" + hashlib.md5(l_password.encode('utf-8')).hexdigest() + ")"
        self.debug(logging.DEBUG, "Trying to login with %s:%s" % \
                (repr(l_user), repr(hashed_pw_string)))
        obj = self.json_obj('user.authenticate', {'user': l_user,
                'password': l_password})
        result = self.do_request(obj)
        self.auth = result['result']

    def test_login(self):
        if self.auth != '':
            obj = self.json_obj('user.checkAuthentication', {'sessionid': self.auth})
            result = self.do_request(obj)

            if not result['result']:
                self.auth = ''
                return False  # auth hash bad
            return True  # auth hash good
        else:
            return False

    def do_request(self, json_obj):
        headers = {'Content-Type': 'application/json-rpc',
                   'User-Agent': 'python/zabbix_api'}

        if self.httpuser:
            self.debug(logging.INFO, "HTTP Auth enabled")
            auth = 'Basic ' + string.strip(base64.encodestring(self.httpuser + ':' + self.httppasswd))
            headers['Authorization'] = auth

        self.r_query.append(str(json_obj))
        self.debug(logging.INFO, "Sending: " + str(json_obj), safe=True)
        self.debug(logging.DEBUG, "Sending headers: " + str(headers))

        request = urllib2.Request(url=self.url, data=json_obj.encode('utf-8'), headers=headers)
        if self.proto == "https":
            https_handler = urllib2.HTTPSHandler(debuglevel=0)
            opener = urllib2.build_opener(https_handler)
        elif self.proto == "http":
            http_handler = urllib2.HTTPHandler(debuglevel=0)
            opener = urllib2.build_opener(http_handler)
        else:
            raise ZabbixAPIException("Unknow protocol %s" % self.proto)

        urllib2.install_opener(opener)
        try:
            response = opener.open(request, timeout=self.timeout)
        except Exception as e:
            raise ZabbixAPIException("Site needs HTTP authentication. Error: "+str(e))
        self.debug(logging.INFO, "Response Code: " + str(response.code))

        # NOTE: Getting a 412 response code means the headers are not in the
        # list of allowed headers.
        if response.code != 200:
            raise ZabbixAPIException("HTTP ERROR %s: %s"
                    % (response.status, response.reason))
        reads = response.read()
        if len(reads) == 0:
            raise ZabbixAPIException("Received zero answer")
        try:
            jobj = json.loads(reads.decode('utf-8'))
        except ValueError as msg:
            print ("unable to decode. returned string: %s" % reads)
            sys.exit(-1)
        self.debug(logging.DEBUG, "Response Body: " + str(jobj))

        self.id += 1

        if 'error' in jobj:  # some exception
            msg = "Error %s: %s, %s while sending %s" % (jobj['error']['code'],
                    jobj['error']['message'], jobj['error']['data'], str(json_obj))
            if re.search(".*already\sexists.*", jobj["error"]["data"], re.I):  # already exists
                raise Already_Exists(msg, jobj['error']['code'])
            else:
                raise ZabbixAPIException(msg, jobj['error']['code'])
        return jobj

    def logged_in(self):
        if self.auth != '':
            return True
        return False

    def api_version(self, **options):
        self.__checkauth__()
        obj = self.do_request(self.json_obj('APIInfo.version', options))
        return obj['result']

    def __checkauth__(self):
        if not self.logged_in():
            raise ZabbixAPIException("Not logged in.")


class ZabbixAPISubClass(ZabbixAPI):
    """ wrapper class to ensure all calls go through the parent object """
    parent = None
    data = None

    def __init__(self, parent, data, **kwargs):
        self._setuplogging()
        self.debug(logging.DEBUG, "Creating %s" % self.__class__.__name__)
        self.data = data
        self.parent = parent

        # Save any extra info passed in
        for key, val in kwargs.items():
            setattr(self, key, val)
            self.debug(logging.WARNING, "Set %s:%s" % (repr(key), repr(val)))

    def __getattr__(self, name):
        if self.data["prefix"] == "configuration" and name == "import_":  # workaround for "import" method
            name = "import"

        def method(*opts):
            return self.universal("%s.%s" % (self.data["prefix"], name), opts[0])
        return method

    def __checkauth__(self):
        self.parent.__checkauth__()

    def do_request(self, req):
        return self.parent.do_request(req)

    def json_obj(self, method, param):
        return self.parent.json_obj(method, param)

    @dojson2
    @checkauth
    def universal(self, **opts):
        return opts



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-13 Erigones s.r.o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""
import logging
from tabulate import tabulate

from ludolph.command import command, parameter_required
from ludolph.plugins.plugin import LudolphPlugin
from ludolph.plugins.zabbix_api import ZabbixAPI, ZabbixAPIException

TIMEOUT = 10
TABLEFMT = 'simple'

logger = logging.getLogger(__name__)

def zabbix_command(f):
    """
    Decorator for executing zabbix API commands, checking zabbix API errors
    and doing a relogin if needed.
    """
    def wrap(obj, msg, *args, **kwargs):
        def api_error(errmsg='Zabbix API not available'):
            # Log and reply with error message
            logger.error(errmsg)
            msg.reply(errmsg).send()
            return None

        # Was never logged in. Repair authentication and restart Ludolph.
        if not obj.zapi.logged_in():
            return api_error()

        try:
            return f(obj, msg, *args, **kwargs)
        except ZabbixAPIException as ex:
            ex = str(ex)
            if ex.find('Not authorized while sending') >= 0:
                # Try to relogin
                try:
                    logger.warning('Zabbix API not logged in (%s). '
                            'Performing Zabbix API relogin.', ex)
                    obj.zapi.auth = '' # Reset auth before relogin
                    obj.zapi.login()
                except ZabbixAPIException as e:
                    # Relogin failed. Repair authentication and restart Ludolph.
                    logger.critical('Zabbix API login error (%s)', e)
                    obj.zapi.auth = '' # logged_in() will always return False
                    return api_error()

                # Relogin successfull, Try to run command
                try:
                    return f(obj, msg, *args, **kwargs)
                except ZabbixAPIException as exc:
                    # API command problem
                    return api_error('Zabbix API error (%s)' % exc)

            else:
                # API command problem
                return api_error('Zabbix API error (%s)' % ex)

    return wrap

class Zabbix(LudolphPlugin):
    """
    Zabbix API connector for LudolphBot.
    """
    zapi = None

    def __init__(self, config, *args, **kwargs):
        """
        Initialize zapi and try to login.
        """
        # HTTP authentication?
        httpuser = None
        httppasswd = None
        if config.has_option('zabbix', 'httpuser'):
            httpuser = config.get('zabbix', 'httpuser')
        if config.has_option('zabbix', 'httppasswd'):
            httppasswd = config.get('zabbix', 'httppasswd')

        self.zapi = ZabbixAPI(server=config.get('zabbix','server'),
                user=httpuser, passwd=httppasswd, timeout=TIMEOUT,
                log_level=logging.getLevelName(config.get('global','loglevel')))

        # Login and save zabbix credentials
        try:
            logger.info('Zabbix API login')
            self.zapi.login(config.get('zabbix', 'username'),
                            config.get('zabbix', 'password'), save=True)
        except ZabbixAPIException as e:
            logger.critical('Zabbix API login error (%s)', e)

    @zabbix_command
    @command
    def zabbix_version(self, msg):
        """
        Show version of Zabbix API.
        """
        return 'Zabbix API version: '+ self.zapi.api_version()

    @zabbix_command
    @command
    def alerts(self, msg):
        """
        List current zabbix alerts.

        Emulates include/blocks.inc.php :: make_latest_issues()
        """
        # get triggers
        options = {
                'groupids': None,
                'hostids': None,
                'monitored': True,
                'maintenance': False,
                'skipDependent': True,
                'filter': {'priority': None, 'value': 1}, # TRIGGER_VALUE_TRUE
                'selectHosts': ['hostid', 'name'],
                'output': ['triggerid', 'value_flags', 'error', 'url', 'expression', 'description', 'priority', 'type'],
                'sortfield': 'lastchange',
                'sortorder': 'DESC', # ZBX_SORT_DOWN
        }
        triggers = self.zapi.trigger.get(options)

        for tnum, trigger in enumerate(triggers):
            # if trigger is lost (broken expression) we skip it
            if not trigger['hosts']:
                del triggers[tnum]
                continue

            host = trigger['hosts'][0]
            trigger['hostid'] = host['hostid']
            trigger['hostname'] = host['name']

            triggers[tnum] = trigger

        # get hosts
        hosts = self.zapi.host.get({
            'hostids': [int(i['hostid']) for i in triggers],
            'output': ['hostid', 'name', 'maintenance_status', 'maintenance_type', 'maintenanceid'],
            'selectInventory': ['hostid'],
            'selectScreens': 'count', #API_OUTPUT_COUNT
            'preservekeys': True,
        })

        # output
        table = []
        for trigger in triggers:
            # get last event
            events = self.zapi.event.get({
                    'output': 'extend',
                    'select_acknowledges': 'extend',
                    'triggerids': trigger['triggerid'],
                    'filter': {
                            'object': 0, # EVENT_OBJECT_TRIGGER
                            'value': 1, # TRIGGER_VALUE_TRUE
                            'value_changed': 1, # TRIGGER_VALUE_CHANGED_YES
                    },
                    'sortfield': ['object', 'objectid', 'eventid'],
                    'sortorder': 'DESC',
                    'limit': 1,
            })

            # Event
            if not events:
                continue # WTF?
            event = events[0]
            eventid = event['eventid']

            # Host and hostname
            host = hosts[trigger['hostid']]
            hostname = host['name']
            if int(host['maintenance_status']):
                hostname += '*' # some kind of maintenance TODO

            # Trigger description
            desc = str(trigger['description'])
            if trigger['error'] or int(trigger['value_flags']):
                desc += '*' # some kind of trigger error TODO

            # Priority
            prio = self.zapi.get_severity(trigger['priority'])

            # Last change and age
            dt = self.zapi.get_datetime(trigger['lastchange'])
            #last = self.zapi.convert_datetime(dt)
            age = self.zapi.get_age(dt)

            # Ack
            if int(event['acknowledged']):
                ack = 'Yes'
            else:
                ack = 'No'

            table.append([eventid, hostname, desc, prio, age, ack])

        out = ''
        if table:
            out = str(tabulate(table, headers=['EventID', 'Host', 'Issue',
                'Severity', 'Age', 'Ack'], tablefmt=TABLEFMT))
            out += '\n\n'

        out += '%d issues are shown.\n%s' % (
            len(triggers), self.zapi.server + '/dashboard.php')

        return out

    @zabbix_command
    @parameter_required(1)
    @command
    def ack(self, msg, eventid):
        """
        Acknowledge event. EventID is a required parameter.
        """
        try:
            eventid = int(eventid)
        except ValueError:
            return 'Integer required'

        self.zapi.event.acknowledge({
            'eventids': [eventid],
            'message': str(msg['from']),
        })

        return 'Event %s acknowledged' % eventid



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-13 Erigones s.r.o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""

class LudolphPlugin(object):
    """
    Ludolph plugin base class.
    """
    def __init__(self, config, *args, **kwargs):
        pass



"""
Ludolph: Monitoring Jabber Bot
Version: 0.3.0
Homepage: https://github.com/erigones/Ludolph
Copyright (C) 2012-13 Erigones s.r.o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""

__version__ = '0.3.0'



"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2012-13 Erigones s.r.o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""
from logging import getLogger

logger = getLogger(__name__)

COMMAND_MAP = {} # command : module
COMMANDS = {} # command : help
USERS = [] # List of users
ADMINS = [] # List of admins

def command(f):
    """
    Decorator for registering available commands.
    """
    # Check if command exists
    if f.__name__ in COMMAND_MAP.keys():
        logger.critical('Command "%s" from plugin "%s" overlaps with existing '
                'command from module "%s"' % (f.__name__, f.__module__,
                    COMMAND_MAP[f.__name__]))
        return None

    # Save in module map
    COMMAND_MAP[f.__name__] = f.__module__

    # Save documentation
    if f.__doc__:
        COMMANDS[f.__name__] = f.__doc__.strip()
    else:
        logger.error('Missing documentation for command "%s"', f.__name__)
        COMMANDS[f.__name__] = ''

    def wrap(obj, msg, *args, **kwargs):
        if not USERS or msg['from'].bare in USERS:
            logger.info('User "%s" requested command "%s"' % (msg['from'],msg['body']))
            # Reply with output of function
            out = f(obj, msg, *args, **kwargs)
            logger.debug('Command output: "%s"', out)
            return msg.reply(out).send()
        else:
            logger.warning('Unauthorized command "%s" from "%s"' % (
                msg['body'], msg['from']))
            msg.reply('Permission denied').send()
            return None

    return wrap

def parameter_required(count=1):
    """
    Decorator for required command parameters.
    """
    def parameter_required_decorator(f):
        def wrap(obj, msg, *args, **kwargs):
            #Try to get command parameter
            params = msg['body'].strip().split()[1:]
            if len(params) == count:
                params.extend(args)
                return f(obj, msg, *params, **kwargs)
            else:
                logger.warning('Missing parameter in command "%s" from user "%s"' % (
                    msg['body'], msg['from']))
                msg.reply('Missing parameter').send()
                return None

        return wrap
    return parameter_required_decorator

def admin_required(f):
    """
    Decorator for admin only commands.
    """
    def wrap(obj, msg, *args, **kwargs):
        if not ADMINS or msg['from'].bare in ADMINS:
            return f(obj, msg, *args, **kwargs)
        else:
            logger.warning('Unauthorized command "%s" from user "%s"' % (
                msg['body'], msg['from']))
            msg.reply('Permission denied').send()
            return None

    return wrap



"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2012-13 Erigones s.r.o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""

import os
import sys
import ssl
import time
import signal
import logging
import subprocess
from sleekxmpp import ClientXMPP
from tabulate import tabulate

# In order to make sure that Unicode is handled properly
# in Python 2.x, reset the default encoding.
if sys.version_info < (3, 0):
    from ConfigParser import RawConfigParser
else:
    from configparser import RawConfigParser

from ludolph.command import ( COMMAND_MAP, COMMANDS, USERS, ADMINS,
        command, parameter_required, admin_required )
from ludolph.__init__ import __doc__ as ABOUT
from ludolph.__init__ import __version__ as VERSION

LOGFORMAT = '%(asctime)s %(levelname)-8s %(name)s: %(message)s'
TABLEFMT = 'simple'

logger = logging.getLogger(__name__)

class LudolphBot(ClientXMPP):
    """
    Ludolph bot.
    """
    command_map = COMMAND_MAP
    commands = COMMANDS
    users = USERS
    admins = ADMINS
    plugins = None

    def __init__(self, config, plugins=None, *args, **kwargs):
        # Initialize the SleeXMPP client
        ClientXMPP.__init__(self,
                config.get('xmpp','username'),
                config.get('xmpp','password'))

        # If you are working with an OpenFire server, you will
        # need to use a different SSL version:
        if config.has_option('xmpp', 'sslv3') and \
            config.getboolean('xmpp', 'sslv3'):
                self.ssl_version = ssl.PROTOCOL_SSLv3

        # Auto-authorize is enabled by default. User subscriptions are
        # controlled by self._handle_new_subscription
        self.auto_authorize = True

        # Rest of the configuration
        self.config = config
        self.pipe_file = config.get('global','pipe_file')

        # Users
        if config.has_option('xmpp', 'users'):
            for i in config.get('xmpp', 'users').strip().split(','):
                i = i.strip()
                if i:
                    self.users.append(i)

        # Admins
        if config.has_option('xmpp', 'admins'):
            for i in config.get('xmpp', 'admins').strip().split(','):
                i = i.strip()
                if i:
                    self.admins.append(i)

        # Admins vs. users
        if self.admins and self.users:
            for i in self.admins:
                if i not in self.users:
                    logger.error('Admin user %s is not specified in users. '
                            'This may lead to unexpected behaviour. ', i)

        # Initialize plugins
        self.plugins = {'__main__': self, 'ludolph.bot': self}
        if plugins:
            for plugin, cls in plugins.items():
                logger.info('Initializing plugin %s', plugin)
                self.plugins[plugin] = cls(config)

        # Register event handlers
        self.add_event_handler('session_start', self.session_start)
        self.add_event_handler('message', self.message)

        # Start the monitoring thread for reading the pipe file
        self._start_thread('mon_thread', self.mon_thread)

    def _handle_new_subscription(self, pres):
        """
        xmpp.auto_authorize is True by default, which is fine. But we want to
        restrict this to users only (if set). We do this by overriding the
        automatic subscription mechanism.
        """
        user = pres['from']

        if not self.users or user in self.users:
            logger.info('Allowing user "%s" to auto subscribe', user)
            return super(LudolphBot, self)._handle_new_subscription(pres)
        else:
            logger.warning('User "%s" is not allowed to susbscribe', user)

    def session_start(self, event):
        """
        Process the session_start event.
        """
        self.send_presence()
        self.get_roster()
        self.roster_cleanup()
        logger.info('Registered commands: %s', ', '.join(self.available_commands()))

    def roster_cleanup(self):
        """
        Remove roster items with none subscription.
        """
        roster = self.client_roster
        logger.info('Current auto_authorize: %s', self.auto_authorize)
        logger.info('Current users: %s', ', '.join(self.users))
        logger.info('Current admins: %s', ', '.join(self.admins))
        logger.info('Current roster: %s', ', '.join(roster.keys()))

        # Remove users with none subscription from roster
        # Also remove users that are not in users setting (if set)
        for i in roster.keys():
            if roster[i]['subscription'] == 'none' or \
                (self.users and i not in self.users):
                    logger.warning('Roster item: %s (%s) - removing!' % (
                        i, roster[i]['subscription']))
                    self.send_presence(pto=i, ptype='unsubscribe')
                    self.del_roster_item(i)
            else:
                logger.info('Roster item: %s (%s) - ok' % (
                    i, roster[i]['subscription']))

    def available_commands(self):
        """
        List of all available bot commands.
        """
        return self.commands.keys()

    def message(self, msg):
        """
        Incoming message handler.
        """
        if msg['type'] in ('chat', 'normal'):
            # Seek received text in available commands
            cmd = msg['body'].split()[0].strip()
            if cmd in self.available_commands():
                # Find and run command
                f = getattr(self.plugins[self.command_map[cmd]], cmd)
                return f(msg)
            else:
                # Send message that command was not understod and what to do
                msg.reply('Sorry, I don\'t understand "%s"\n'
                        'Please type "help" for more info' % msg['body']).send()

    def shutdown(self, signalnum, handler):
        """
        Shutdown signal handler.
        """
        logger.info('Requested shutdown (%s)', signalnum)

        return self.abort()

    def mon_thread(self):
        """
        Processing input from the monitoring pipe file.
        """
        with os.fdopen(os.open(self.pipe_file, os.O_RDONLY|os.O_NONBLOCK)) as fifo:
            logger.info('Processing input from monitoring pipe file')
            while not self.stop.is_set():
                line = fifo.readline().strip()
                if line:
                    data = line.split(';', 1)
                    if len(data) == 2:
                        logger.info('Sending monitoring message to %s', data[0])
                        logger.debug('\twith body: "%s"', data[1])
                        self.send_message(mto=data[0], mbody=data[1],
                                mtype='normal')
                    else:
                        logger.warning('Bad message format ("%s")', line)

                time.sleep(1)

                if self.stop.is_set():
                    self._end_thread('mon_thread', early=True)
                    return

    @command
    def help(self, msg):
        """
        Show this help.
        """
        # Global help or command help?
        cmdline = msg['body'].strip().split()
        if len(cmdline) > 1 and cmdline[1] in self.available_commands():
            out = ['* ' + cmdline[1], '', self.commands[cmdline[1]]]
        else:
            out = ['List of available Ludolph commands:']

            for cmd, info in self.commands.items():
                try:
                    # First line of __doc__
                    desc = info.split('\n')[0]
                    # Lowercase first char and remove trailing dot
                    desc = desc[0].lower() + desc[1:].rstrip('.')
                except IndexError:
                    desc = ''
                # Append line of command + description
                out.append('\t* %s - %s' % (cmd, desc))

        return '\n'.join(out)

    @command
    def version(self, msg):
        """
        Display Ludolph version.
        """
        return 'Version: '+ VERSION

    @command
    def about(self, msg):
        """
        Details about this project.
        """
        return ABOUT.strip()

    @admin_required
    @command
    def roster_list(self, msg):
        """
        List of users on Ludolph's roster (admin only).
        """
        roster = self.client_roster
        out = []

        for i in roster.keys():
            out.append((str(i), roster[i]['subscription'],))

        return str(tabulate(out, headers=['JID', 'subscription'], tablefmt=TABLEFMT))

    @admin_required
    @parameter_required(1)
    @command
    def roster_remove(self, msg, user):
        """
        Remove user from Ludolph's roster (admin only).
        """
        if user in self.client_roster.keys():
            self.send_presence(pto=user, ptype='unsubscribe')
            self.del_roster_item(user)
            return 'User '+ user +' removed from roster'
        else:
            return 'User '+ user +' cannot be removed from roster'

    @command
    def uptime(self, msg):
        """
        Show server uptime.
        """
        cmd = subprocess.Popen(['uptime'], stdout=subprocess.PIPE)

        return cmd.communicate()[0]


def daemonize():
    """
    http://code.activestate.com/recipes/278731-creating-a-daemon-the-python-way/
    http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
    """
    try:
        pid = os.fork() # Fork #1
        if pid > 0:
            sys.exit(0) # Exit first parent
    except OSError as e:
        sys.stderr.write('Fork #1 failed: %d (%s)\n' % (e.errno, e.strerror))
        sys.exit(1)

    # The first child. Decouple from parent environment
    # Become session leader of this new session.
    # Also be guaranteed not to have a controlling terminal
    os.chdir('/')
    os.setsid()
    os.umask(0)

    try:
        pid = os.fork() # Fork #2
        if pid > 0:
            sys.exit(0) # Exit from second parent
    except OSError as e:
        sys.stderr.write('Fork #2 failed: %d (%s)\n' % (e.errno, e.strerror))
        sys.exit(1)

    # Close all open file descriptors
    import resource # Resource usage information
    maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if (maxfd == resource.RLIM_INFINITY):
        maxfd = 1024

    # Iterate through and close all file descriptors
    for fd in range(0, maxfd):
        try:
            os.close(fd)
        except OSError: # ERROR, fd wasn't open (ignored)
            pass

    # Redirect standard file descriptors to /dev/null
    sys.stdout.flush()
    sys.stderr.flush()
    si = open(os.devnull, 'r')
    so = open(os.devnull, 'a+')
    se = open(os.devnull, 'a+')
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    return(0)

def main():
    """
    Start the daemon.
    """
    ret = 0
    cfg = 'ludolph.cfg'
    cfg_fp = None
    cfg_lo = ((os.path.expanduser('~'), '.'+ cfg),
            (sys.prefix, 'etc', cfg), ('/etc', cfg))
    config = RawConfigParser()
    config_base_sections = ('global', 'xmpp')

    # Try to read config file from ~/.ludolph.cfg or /etc/ludolph.cfg
    for i in cfg_lo:
            try:
                cfg_fp = open(os.path.join(*i))
            except IOError:
                continue
            else:
                break

    if cfg_fp:
        config.readfp(cfg_fp)
    else:
        sys.stderr.write("""\nLudolph can't start!\n
You need to create a config file in one this locations: \n%s\n
You can rename ludolph.cfg.example and update the required variables.
The example file is located in: %s\n\n""" % (
        '\n'.join([os.path.join(*i) for i in cfg_lo]),
        os.path.dirname(os.path.abspath(__file__))))
        sys.exit(1)

    # Prepare logging configuration
    logconfig = {
        'level': logging.getLevelName(config.get('global','loglevel')),
        'format': LOGFORMAT,
    }

    if config.has_option('global', 'logfile'):
        logfile = config.get('global','logfile').strip()
        if logfile:
            logconfig['filename'] = logfile

    # Daemonize
    if config.has_option('global', 'daemon'):
        if config.getboolean('global', 'daemon'):
            ret = daemonize()
            # Save pid file
            try:
                with open(config.get('global', 'pidfile'), 'w') as fp:
                    fp.write('%s' % os.getpid())
            except Exception as ex:
                # Setup logging just to show this error
                logging.basicConfig(**logconfig)
                logger.critical('Could not write to pidfile (%s)\n', ex)
                sys.exit(1)

    # Setup logging
    logging.basicConfig(**logconfig)

    # All exceptions will be logged
    def log_except_hook(*exc_info):
        logger.critical('Unhandled exception!', exc_info=exc_info)
        #sys.exit(99)
    sys.excepthook = log_except_hook

    # Default configuration
    pipe_file = config.get('global','pipe_file')
    pipe_mode = '0600'
    use_tls = True
    use_ssl = False
    address = []

    # Starting
    logger.info('Starting Ludolph %s', VERSION)
    logger.info('Loaded configuration from %s', cfg_fp.name)

    # Load plugins
    plugins = {}
    for plugin in config.sections():
        plugin = plugin.lower().strip()
        if plugin in config_base_sections:
            continue
        logger.info('Loading plugin: %s', plugin)
        try:
            clsname = plugin[0].upper() + plugin[1:]
            modname = 'ludolph.plugins.'+ plugin
            module = __import__(modname, fromlist=[clsname])
            plugins[modname] = getattr(module, clsname)
        except Exception as ex:
            logger.critical('Could not load plugin %s', plugin)
            logger.exception(ex)

    # XMPP connection settings
    if config.has_option('xmpp', 'host'):
        address = [config.get('xmpp', 'host'), '5222']
        if config.has_option('xmpp', 'port'):
            address[1] = config.get('xmpp', 'port')
        logger.info('Connecting to jabber server %s', ':'.join(address))
    else:
        logger.info('Using DNS SRV lookup to find jabber server')

    if config.has_option('xmpp', 'tls'):
        use_tls = config.getboolean('xmpp', 'tls')

    if config.has_option('xmpp', 'ssl'):
        use_ssl = config.getboolean('xmpp', 'ssl')

    # Create pipe file with desired permissions
    if config.has_option('global', 'pipe_mode'):
        pipe_mode = config.get('global', 'pipe_mode')

    logger.info('Creating pipe file %s', pipe_file)
    try:
        os.remove(pipe_file)
    except:
        pass

    os.mkfifo(pipe_file, int(pipe_mode, 8))

    # Here we go
    try:
        xmpp = LudolphBot(config, plugins=plugins)
        signal.signal(signal.SIGINT, xmpp.shutdown)
        signal.signal(signal.SIGTERM, xmpp.shutdown)
        if xmpp.connect(tuple(address), use_tls=use_tls, use_ssl=use_ssl):
            xmpp.process(block=True)
            sys.exit(ret)
        else:
            logger.error('Ludolph is unable to connect to jabber server')
            sys.exit(2)
    finally:
        # Cleanup
        logger.info('Removing pipe file %s', pipe_file)
        os.remove(pipe_file)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-13 Erigones s.r.o.
# All Rights Reserved
#
# This software is licensed as described in the README.rst and LICENSE
# files, which you should have received as part of this distribution.

import codecs
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from ludolph.__init__ import __version__ as VERSION

DESCRIPTION = 'Monitoring Jabber Bot'

with codecs.open('README.rst', 'r', encoding='UTF-8') as readme:
    LONG_DESCRIPTION = ''.join(readme)

CLASSIFIERS = [
    'Environment :: Console',
    'Intended Audience :: System Administrators',
    'Operating System :: Unix',
    'Operating System :: POSIX :: Linux',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Topic :: Communications :: Chat',
    'Topic :: Utilities'
]

packages = [
    'ludolph',
]

setup(
    name = 'ludolph',
    version = VERSION,
    description = DESCRIPTION,
    long_description = LONG_DESCRIPTION,
    author = 'Erigones',
    author_email = 'erigones [at] erigones.com',
    url = 'https://github.com/erigones/Ludolph/',
    license = 'GPLv3',
    packages = packages,
    scripts = ['bin/ludolph'],
    install_requires = ['sleekxmpp>=1.1.11', 'dnspython', 'tabulate'],
    platforms = 'Linux',
    classifiers = CLASSIFIERS,
    include_package_data = True
)



"""
Copyright (C) 2012-13 Erigones s.r.o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""
import subprocess

def available_commands():
    available_commands = ['who', 'uptime']
    return available_commands

#def enabled_commands():

def execute_command(cmd):
    if cmd in available_commands():
        return True, str(subprocess.check_output(cmd))
    else:
        return False, 'Command '+ cmd +' was not recognized.'
